local _G						= getfenv(0)

local ipairs					= _G.ipairs
local next						= _G.next
local pairs						= _G.pairs
local select					= _G.select
local strsplit					= _G.strsplit
local unpack					= _G.unpack

local table_concat				= _G.table.concat
local table_insert				= _G.table.insert
local table_remove				= _G.table.remove
local table_sort				= _G.table.sort
local table_wipe				= _G.table.wipe

local BNConnected				= _G.BNConnected
local BNFeaturesEnabled			= _G.BNFeaturesEnabled
local BNGetFriendInfo			= _G.BNGetFriendInfo
local BNGetNumFriends			= _G.BNGetNumFriends
local BNGetToonInfo				= _G.BNGetToonInfo
local CanCooperateWithToon		= _G.CanCooperateWithToon
local GetFriendInfo				= _G.GetFriendInfo
local GetNumFriends				= _G.GetNumFriends
local GetQuestDifficultyColor	= _G.GetQuestDifficultyColor
local GetTime					= _G.GetTime
local HasTravelPass				= _G.HasTravelPass
local ShowFriends				= _G.ShowFriends
local UnitInParty				= _G.UnitInParty
local UnitInRaid				= _G.UnitInRaid

local classLookup = {}
for k, v in pairs(_G.LOCALIZED_CLASS_NAMES_MALE) do
	classLookup[v] = k
end
for k, v in pairs(_G.LOCALIZED_CLASS_NAMES_FEMALE) do
	classLookup[v] = k
end

local ElkFriends = _G.LibStub("AceAddon-3.0"):NewAddon("ElkFriends", "AceBucket-3.0", "AceConsole-3.0", "AceEvent-3.0")

local LDB = _G.LibStub("LibDataBroker-1.1")
local LDBIcon = _G.LibStub("LibDBIcon-1.0")
local LibQTip = _G.LibStub("LibQTip-1.0")

local DBdefaults = {
	profile = {
		minimap = {
			hide = false,
		},
		sorting = "ZONE:LEVEL:NAME:CLASS:NOTE",
		filterClass = {
			DEATHKNIGHT = true,
			DRUID = true,
			HUNTER = true,
			MAGE = true,
			MONK = true,
			PALADIN = true,
			PRIEST = true,
			ROGUE = true,
			SHAMAN = true,
			WARLOCK = true,
			WARRIOR = true,
		},
	},
}


local L = _G.setmetatable({
	TEXT_UPDATING = "Updating...",
	TOOLTIP_COLUMN_HEADER_NAME = "Name",
	TOOLTIP_COLUMN_HEADER_LEVEL = "Level",
	TOOLTIP_COLUMN_HEADER_ZONE = "Zone",
	TOOLTIP_COLUMN_HEADER_NOTES = "Notes",
	TOOLTIP_TEXT_NOFRIENDSONLINE = "There are no known friends online.",
}, {__index = function(self, key) return key end })

-- -----
-- Roster Updater
-- -----
local nextShowFriends = 0
local updateFrame = _G.CreateFrame("frame")
updateFrame:Hide()
updateFrame:SetScript("OnUpdate", function(self, elapsed)
	if (GetTime() < nextShowFriends) then return end
	ShowFriends()
end)

_G.hooksecurefunc("ShowFriends", function()
	nextShowFriends = GetTime() + 15
end)

-- -----
-- player table cache
-- -----
local tablecache = {}
local function recyclePlayerTable(t)
	while #t > 0 do
		table_insert(tablecache, table_remove(t))
	end
	return t
end
local function getPlayerTable(...)
	local t
	if #tablecache > 0 then
		t = table_remove(tablecache)
	else
		t = {}
	end
	for i = 1, select('#', ...) do
		t[i] = select(i, ...)
	end
	return t
end

-- -----
local function do_OnEnter(frame)
	local tooltip = LibQTip:Acquire("ElkFriendsTip")
	tooltip:SmartAnchorTo(frame)
	tooltip:SetAutoHideDelay(0.1, frame)
	tooltip:EnableMouse(true)
	ElkFriends:UpdateTooltip()
	tooltip:Show()
	updateFrame:Show()
end

local function do_OnLeave()
	updateFrame:Hide()
end

local function do_OnClick(frame, button)
	if button == "RightButton" then
		_G.InterfaceOptionsFrame_OpenToCategory("ElkFriends")
	else
		_G.ToggleFriendsFrame(1)
	end
end

function ElkFriends:OnInitialize()
	self.players = {}
	self.playersBN = {}

	self.dbo = _G.LibStub("LibDataBroker-1.1"):NewDataObject("ElkFriends", {
			type = "data source",
			text = L["TEXT_UPDATING"],
			icon = [[Interface\Addons\ElkFriends\ElkFriends]],
			OnEnter = do_OnEnter,
			OnLeave = do_OnLeave,
			OnClick = do_OnClick,
		})

	self.db = _G.LibStub("AceDB-3.0"):New("ElkFriendsDB", DBdefaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	self:LoadSort()

	LDBIcon:Register("ElkFriends", self.dbo, self.db.profile.minimap)
end

function ElkFriends:OnEnable()
	self:RegisterBucketEvent({"FRIENDLIST_UPDATE", "BN_FRIEND_TOON_ONLINE", "BN_FRIEND_TOON_OFFLINE"}, .5, "UpdateInfo")
	self:UpdateInfo()
end

function ElkFriends:RefreshConfig()
	-- minimap icon
	LDBIcon:Refresh("ElkFriends", self.db.profile.minimap)
	
	self:LoadSort()
	self:UpdateData()
	self:UpdateText()
	self:UpdateTooltip()
end

function ElkFriends:UpdateInfo()
	nextShowFriends = GetTime() + 15

	self:UpdateData()

	-- update the LDB text
	self:UpdateText()
	
	-- update the tooltip
	self:UpdateTooltip()
end

-- -----
-- FILTERING
-- -----
function ElkFriends:checkFilter(classFileName, level, zone)
	return true
end

-- -----
-- SORTING
-- -----
local SORT_CONST = {
	[1] = "NAME",			["NAME"] = 1,			[-1] = "-NAME",			["-NAME"] = -1,
	[2] = "LEVEL",			["LEVEL"] = 2,			[-2] = "-LEVEL",		["-LEVEL"] = -2,
	[3] = "CLASS",			["CLASS"] = 3,			[-3] = "-CLASS",		["-CLASS"] = -3,
	[5] = "ZONE",			["ZONE"] = 5,			[-5] = "-ZONE",			["-ZONE"] = -5,
	[6] = "NOTE",			["NOTE"] = 6,			[-6] = "-NOTE",			["-NOTE"] = -6,
}

local sortOrder = {SORT_CONST.ZONE, SORT_CONST.LEVEL, SORT_CONST.NAME, SORT_CONST.CLASS, SORT_CONST.NOTE}
local sortIndex = 1
local function sorter(a, b)
	local sortField = sortOrder[sortIndex]
	if (a == b or not sortField) then
		sortIndex = 1
		return false
	end
	local asc
	if (sortField > 0) then
		asc = true
	else
		sortField = -sortField
		asc = false
	end
	local aVal = a[sortField]
	local bVal = b[sortField]
	-- if (type(aVal) ~= type(bVal)) then
		-- print(aVal, bVal, type(aVal), type(bVal))
	-- end
	if (bVal and (not aVal or aVal < bVal)) then
		sortIndex = 1
		return asc
	elseif (aVal and (not bVal or aVal > bVal)) then
		sortIndex = 1
		return not asc
	else
		sortIndex = sortIndex + 1
		return sorter(a, b)
	end
end

function ElkFriends:LoadSort()
	local sortString = self.db.profile.sorting
	table_wipe(sortOrder)
	for key in sortString:gmatch("(%-?[A-Z]+)") do
		local sortField = SORT_CONST[key]
		if (sortField) then
			table_insert(sortOrder, sortField)
		end
	end
end

function ElkFriends:SaveSort()
	local sortString = {}
	for _, v in ipairs(sortOrder) do
		table_insert(sortString, SORT_CONST[v])
	end
	self.db.profile.sorting = table_concat(sortString, ":")
end

function ElkFriends:UpdateData() --TODO
	local numBNetTotal, numBNetOnline = 0, 0
	local numWoWTotal, numWoWOnline = GetNumFriends()

	local players = recyclePlayerTable(self.players)
	local playersBN = recyclePlayerTable(self.playersBN)
	local playersShown = 0
	local playersOnline = 0

	local name, level, class, classFileName, zone, online, status, note
	for i = 1, numWoWOnline do
		name, level, class, zone, online, status, note = GetFriendInfo(i)
		if (online) then
			playersOnline = playersOnline + 1
			classFileName = classLookup[class] or "PRIEST" -- PRIEST is white
			if (not note or note == "") then note = false end
			if self:checkFilter(classFileName, level, zone) then
				playersShown = playersShown + 1
				table_insert(players, getPlayerTable(
					name or _G.UNKNOWN,
					level or -1,
					class or _G.UNKNOWN,
					classFileName,
					zone,
					note,
					status
				))
			end
		end
	end
	
	if BNFeaturesEnabled() and BNConnected() then
		numBNetTotal, numBNetOnline = BNGetNumFriends()
		local hasTravelPass = HasTravelPass()
		local presenceID, presenceName, battleTag, isBattleTagPresence, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, messageText, noteText, isRIDFriend, messageTime, canSoR
		local hasFocus, toonName, client, realmName, realmID, faction, race, class, classFileName, guild, zoneName, level, gameText, broadcastText, broadcastTime, canSoR, toonID
		for i = 1, numBNetOnline do
			presenceID, presenceName, battleTag, isBattleTagPresence, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, messageText, noteText, isRIDFriend, messageTime, canSoR = BNGetFriendInfo(i)
			if (isOnline) then
				playersOnline = playersOnline + 1
				if toonID then
					hasFocus, toonName, client, realmName, realmID, faction, race, class, guild, zone, level, gameText, broadcastText, broadcastTime, canSoR, toonID = BNGetToonInfo(toonID)
				else
					hasFocus, toonName, client, realmName, realmID, faction, race, class, guild, zone, level, gameText, broadcastText, broadcastTime, canSoR, toonID = BNGetToonInfo(presenceID)
				end
				-- add to BNet list
				if (not noteText or noteText == "") then noteText = false end
				table_insert(playersBN, getPlayerTable(
					presenceID,
					presenceName or _G.UNKNOWN,
					client,
					toonName,
					gameText,
					noteText,
					isAFK and _G.CHAT_FLAG_AFK or isDND and _G.CHAT_FLAG_DND or ""
				))
				
				if (client == _G.BNET_CLIENT_WOW) then
					level = level and level + 0 -- BNet level is a string						
					if (CanCooperateWithToon(toonID, hasTravelPass)) then
						-- add toon to friends list
						classFileName = classLookup[class] or "PRIEST" -- PRIEST is white
						if self:checkFilter(classFileName, level, zone) then
							playersShown = playersShown + 1
							table_insert(players, getPlayerTable(
								toonName.."-"..realmName or _G.UNKNOWN,
								level or -1,
								class or _G.UNKNOWN,
								classFileName,
								zone,
								noteText,
								isAFK and _G.CHAT_FLAG_AFK or isDND and _G.CHAT_FLAG_DND or ""
							))
						end
					end
				end
			end
		end
	end
	
	table_sort(players, sorter)
	
	self.players = players
	self.playersShown = playersShown
	self.playersOnline = playersOnline
	self.playersTotal = numBNetTotal + numWoWTotal
end

function ElkFriends:UpdateText()
	self.dbo.text = self.playersOnline.."/"..self.playersTotal
end

-- -----
-- TOOLTIP
-- -----
local function OnHeaderClick(frame, sortString, button)
	local leftSort, rightSort = sortString:match("([A-Z]*):?([A-Z]*)")
	local targetSort = (button == "RightButton") and rightSort or leftSort
	if (targetSort == "") then
		return
	else
		targetSort = SORT_CONST[targetSort]
	end
	local mainSort = sortOrder[1]
	if (mainSort == targetSort) then
		sortOrder[1] = -targetSort
	elseif (mainSort == -targetSort) then
		sortOrder[1] = targetSort
	else
		for k, v in ipairs(sortOrder) do
			if (v == targetSort or v == -targetSort) then
				table_remove(sortOrder, k)
				break
			end
		end
		table_insert(sortOrder, 1, targetSort)
	end

	ElkFriends:SaveSort()
	table_sort(ElkFriends.players, sorter)
	ElkFriends:UpdateTooltip()
end

local function OnNameClick(frame, name, button)
	if not name then return end
	if _G.IsAltKeyDown() then
		_G.InviteUnit(name)
	else
--~ 		_G.SetItemRef("player:"..name, "", "LeftButton")
		_G.SetItemRef("player:"..name, "", button)
	end
end

local function OnBNetClick(frame, name, button)
--~ 	_G.SetItemRef("BNplayer:"..name, "", "LeftButton")
	_G.SetItemRef("BNplayer:"..name, "", button)
end

local function RGB2HEX(red, green, blue)
	return ("|cff%02x%02x%02x"):format(red * 255, green * 255, blue * 255)
end

local statusMap = {
	[""] =				"",
	[_G.CHAT_FLAG_AFK] =	"|TInterface\\FriendsFrame\\StatusIcon-Away:14:14:0:0:16:16:0:16:0:16|t",
	[_G.CHAT_FLAG_DND] =	"|TInterface\\FriendsFrame\\StatusIcon-DnD:14:14:0:0:16:16:0:16:0:16|t",
}
local function getSortArrow(sorting)
	if (sortOrder[1] == SORT_CONST[sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:-1:16:8:0:8:0:8|t"
	elseif (sortOrder[1] == SORT_CONST["-"..sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:0:16:8:0:8:8:0|t"
	elseif (sortOrder[2] == SORT_CONST[sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:-1:16:8:0:8:0:8:127:127:127|t"
	elseif (sortOrder[2] == SORT_CONST["-"..sorting]) then
		return " |TInterface\\Buttons\\UI-SortArrow:8:8:0:0:16:8:0:8:8:0:127:127:127|t"
	else
		return ""
	end
end

function ElkFriends:UpdateTooltip() --TODO
	if not LibQTip:IsAcquired("ElkFriendsTip") then
		updateFrame:Hide()
		return
	end
	local tooltip = LibQTip:Acquire("ElkFriendsTip")
	tooltip:Clear()

	local lineNum
	tooltip:SetColumnLayout(5, "CENTER", "LEFT", "CENTER", "CENTER", "RIGHT")

	lineNum = tooltip:AddHeader()
	tooltip:SetCell(lineNum, 1, "|cfffed100Elk|cffffffffFriends", tooltip:GetHeaderFont(), "CENTER", 5)

	lineNum = tooltip:AddLine()
	tooltip:SetCell(lineNum, 2, "|cffffd200"..getSortArrow("CLASS")..L["TOOLTIP_COLUMN_HEADER_NAME"]..getSortArrow("NAME"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 2, "OnMouseUp", OnHeaderClick, "NAME:CLASS")
	tooltip:SetCell(lineNum, 3, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_LEVEL"]..getSortArrow("LEVEL"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 3, "OnMouseUp", OnHeaderClick, "LEVEL")
	tooltip:SetCell(lineNum, 4, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_ZONE"]..getSortArrow("ZONE"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 4, "OnMouseUp", OnHeaderClick, "ZONE")
	tooltip:SetCell(lineNum, 5, "|cffffd200"..L["TOOLTIP_COLUMN_HEADER_NOTES"]..getSortArrow("NOTE"), nil, "CENTER")
	tooltip:SetCellScript(lineNum, 5, "OnMouseUp", OnHeaderClick, "NOTE")
	lineNum = tooltip:AddSeparator(2)
	
	for _, player in ipairs(self.players) do
		local name, level, class, classFileName, zone, note, status = unpack(player)
		local pureName = strsplit("@", name)
		local sameGroup = UnitInParty(pureName) or UnitInRaid(pureName)
		local statusIcon = statusMap[status]
		local levelcolor = GetQuestDifficultyColor(level)
		levelcolor = RGB2HEX(levelcolor.r, levelcolor.g, levelcolor.b)
		local tipnote = note and "["..note.."]" or "-"
		lineNum = tooltip:AddLine()
		tooltip:SetCell(lineNum, 1, sameGroup and "|TInterface\\Buttons\\UI-CheckBox-Check:14:14|t" or "")	-- same group?
		tooltip:SetCell(lineNum, 2, statusIcon.."|c".._G.RAID_CLASS_COLORS[classFileName].colorStr..name)	-- status / name
		tooltip:SetCell(lineNum, 3, levelcolor..level)														-- level
		tooltip:SetCell(lineNum, 4, "|cffffffff"..zone)																	-- zone
		tooltip:SetCell(lineNum, 5, "|cffffffff"..tipnote)																-- notes
		tooltip:SetLineScript(lineNum, "OnMouseUp", OnNameClick, name)
	end

	lineNum = tooltip:AddSeparator(2)

	for _, player in ipairs(self.playersBN) do
		local presenceID, name, client, toonName, gameText, note, status = unpack(player)
		local statusIcon = statusMap[status]
		local tipnote = note and "["..note.."]" or "-"
		lineNum = tooltip:AddLine()
		tooltip:SetCell(lineNum, 1, BNet_GetClientEmbeddedTexture(client, 14))								-- client icon
		tooltip:SetCell(lineNum, 2, statusIcon.."|cffffffff"..name.." ("..toonName.." - "..gameText..")", nil, nil, 3)	-- status / name
		tooltip:SetCell(lineNum, 5, "|cffffffff"..tipnote)																-- notes
		tooltip:SetLineScript(lineNum, "OnMouseUp", OnBNetClick, name..":"..presenceID)
	end
	tooltip:UpdateScrolling()
end